==> sslf-0.1.7-py2.py3-none-any.whl: sslf/sslf.py <==
"""
The main module of sslf (simple spectral line finder).
The provided Spectrum class is intended to facilitate all functionality.
"""

# Python 2 and 3 compatibility
from future.builtins import range
from future.utils import raise_with_traceback

import copy
import logging

import numpy as np
import numpy.ma as ma
from scipy import signal


logger = logging.getLogger(__name__)


def find_background_rms(array, num_chunks, use_chunks):
    """
    In order to obtain a better estimate of a spectrum's RMS, break the input
    array into evenly sized chunks (or as even as possible), sort the chunks
    by RMS, then return the average of some of these chunks as the true RMS.

    Parameters
    ----------
    array : array_like
        The 1-dimensional spectrum we wish to determine the RMS for.

    num_chunks : int, optional
        Split `array` into this many chunks for RMS calculation.

    use_chunks : int, optional
        Use this many chunks from `array` for RMS calculation.
    """

    chunks = np.array_split(array, num_chunks)
    if isinstance(array, ma.core.MaskedArray):
        # The following requires np.errstate, because numpy.ma apparently throws
        # floating point errors erroneously.
        # https://github.com/numpy/numpy/issues/4895
        with np.errstate(under="ignore"):
            sorted_by_rms = sorted([np.std(x.data) for x in chunks])
    else:
        sorted_by_rms = sorted([np.std(x) for x in chunks])
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("rms chunks = %s", sorted_by_rms)
    mean = np.mean(sorted_by_rms[:use_chunks])
    if mean == 0:
        raise_with_traceback(ValueError("RMS is 0, which cannot be used."))
    return mean


def _blank_spectrum_part(spectrum, point, radius, value=0):
    lower = max([0, point+int(-radius)])
    upper = min([len(spectrum), point+int(radius)])
    spectrum[lower:upper] = value


class _Peak(object):
    def __init__(self, channel, snr, width):
        self.channel = channel
        self.snr = snr
        self.width = width


class Spectrum(object):
    def __init__(self, spectrum, vel=None, num_chunks=5, use_chunks=3):
        """
        Provide a spectrum to find lines on, and/or remove the bandpass from.
        The optional vel parameter essentially provides an "x-axis" to the
        data, such that peak positions can be determined in terms of this
        axis rather than merely the channel position in the spectrum.
        Note that any NaN values in the spectrum are filtered.

        Parameters
        ----------
        spectrum : array_like
            The 1-dimensional spectrum we will performing line-finding on.

        vel : array_like, optional
            Supply an equally-sized array to `spectrum` to have `sslf` report
            the velocities (or some other arbitrary interpretation of the x-axis
            of `spectrum`) of found lines.

        num_chunks : int, optional
            When calculating the RMS of `spectrum`, break `spectrum` into
            equally-sized chunks according to `num_chunks`.

        use_chunks : int, optional
            When calculating the RMS of `spectrum`, use this many equally-sized
            chunks from when `spectrum` was split according to `num_chunks`.
        """

        # Make sure the data are in numpy arrays.
        if isinstance(spectrum, list):
            spectrum = np.array(spectrum)
        if isinstance(vel, list):
            vel = np.array(vel)

        # Filter any NaNs, scipy doesn't like them.
        if np.any(np.isnan(spectrum)):
            nan_indices = np.isnan(spectrum)
            self.original = spectrum[~nan_indices]
            if vel is not None:
                self.vel = vel[~nan_indices]
            else:
                self.vel = None
        else:
            self.original = spectrum
            self.vel = vel

        self.rms = find_background_rms(spectrum, num_chunks=num_chunks, use_chunks=use_chunks)


    def find_cwt_peaks(self, scales, snr=6.5, wavelet=signal.ricker,
                       num_chunks=5, use_chunks=3):
        """
        From the input spectrum (and a range of scales to search):
        - perform a continuous wavelet transformation (CWT)
        - find a significant peak in the CWT matrix
        - mask this peak in wavelet space, for all scales
        - loop from step 2, until no significant peaks remain
        - return the list of peaks

        In my experience, an SNR of 6.5 is a good compromise for reducing the number
        of false positives found while reliably finding real, significant peaks.

        It may be worthwhile to smooth the spectrum before performing the CWT.

        Parameters
        ----------
        scales : array_like
            The channel-widths to use when performing the CWT, and thus the
            channel-widths we use when searching for spectral lines.

        snr : float, optional
            The significance threshold of any found spectral line.

        wavelet : function
            The wavelet to use when performing the CWT. The wavelet used should
            have a similar profile to any spectral lines that are being
            searched.

        num_chunks : int, optional
            When calculating the RMS of `spectrum`, break `spectrum` into
            equally-sized chunks according to `num_chunks`.

        use_chunks : int, optional
            When calculating the RMS of `spectrum`, use this many equally-sized
            chunks from when `spectrum` was split according to `num_chunks`.
        """

        peaks = []
        cwt_mat = signal.cwt(self.original, wavelet, scales)
        cwt_mat = ma.array(cwt_mat)
        spectrum_length = cwt_mat.shape[1]

        while True:
            peak_pixel = cwt_mat.argmax()
            i, peak_channel = np.unravel_index(peak_pixel, cwt_mat.shape)
            peak = cwt_mat[i, peak_channel]
            rms = find_background_rms(cwt_mat[i], num_chunks=num_chunks, use_chunks=use_chunks)
            sig = peak/rms
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Biggest peak at channel %s, scale %s, rms = %s",
                             peak_channel, scales[i], rms)

            # If this maximum is not significant, we're done.
            if sig < snr:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Peak is not significant (%s < %s); finishing", sig, snr)
                break
            # Otherwise, blank this line across all scales.
            else:
                for j, s in enumerate(scales):
                    # If the line is too close to the edge,
                    # cap the mask at the edge.
                    lower = max([0, peak_channel - 2*s])
                    upper = min([spectrum_length, peak_channel + 2*s])
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Blanked channels %s to %s (scale = %s)",
                                     lower, upper, s)
                    cwt_mat[j, lower:upper] = ma.masked
                peaks.append(_Peak(peak_channel, sig, scales[i]))

        self.channel_peaks = [p.channel for p in peaks]
        self.peak_snrs = [p.snr for p in peaks]
        self.peak_widths = [p.width for p in peaks]
        if self.vel is not None:
            self.vel_peaks = [self.vel[p.channel] for p in peaks]
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Channel peaks: %s", self.channel_peaks)
            logger.debug("Peak SNRs: %s", self.peak_snrs)
            logger.debug("Peak widths: %s", self.peak_widths)
            if self.vel is not None:
                logger.debug("Velocity peaks: %s", self.vel_peaks)


    def vel_peaks2chan_peaks(self):
        """
        This function is useful for when you know the velocities of the spectral lines,
        and need to determine the relevant channels before subtracting the bandpass.
        """
        self.channel_peaks = []
        for vp in self.vel_peaks:
            self.channel_peaks.append(np.abs(self.vel-vp).argmin())
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Channel peaks: %s", self.channel_peaks)


    def subtract_bandpass(self,
                          window_length=151,
                          poly_order=1,
                          blank_spectrum_width=1.4,
                          allowable_peak_gap=10):
        """
        Flag the locations of any lines, and subtract the non-zero bandpass
        everywhere else. Provide the flattened spectrum in self.modified.

        Parameters
        ----------
        window_length : int, optional
            The window_length to be used in scipy's savgol_filter. This should
            be bigger than the expected width of any spectral lines, but not
            so big that it is comparable to the length of the spectrum. In a
            sense, it specifies how far to look ahead and behind every channel
            when considering bandpass shape. If this is too small, it will
            behave more like a low-pass filter than a high-pass filter, when we
            probably want it to be more on the high-pass side.

        poly_order : int, optional
            The order of the polynomial fitting to be done over the spectral
            lines and with the savgol_filter. 1 (linear) is generally sensible,
            especially as higher orders are susceptible to Runge's phenomenon.

        blank_spectrum_width : float, optional
            The amount of spectrum to filter on either side of a detected
            spectral line. This is in an effort to preserve real signal
            associated with the spectral line that is buried in the noise.

        allowable_peak_gap : int, optional
            The minimum number of channels between any two spectral lines before
            sslf considers them to be the same for the purposes of filtering.
            This avoids using noisy channels between lines for interpolation,
            and getting a poor subtraction. Currently disabled.
        """

        mask = np.zeros(len(self.original))

        for i, p in enumerate(self.channel_peaks):
            width = self.peak_widths[i] * blank_spectrum_width

            # Blank the lines, fitting the bandpass around them.
            _blank_spectrum_part(mask, p, radius=width, value=1)

        self.filtered = copy.copy(self.original)

        # Find where lines start and end.
        edges = []
        if mask[0] == 1:
            edges.append(0)
        for e in np.where(np.diff(mask))[0]:
            edges.append(e)
        if mask[-1] == 1:
            edges.append(len(mask)-1)

        # Interpolate between gaps in the spectrum.
        for i in range(len(edges)//2):
            e1, e2 = edges[2*i], edges[2*i+1]
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Interpolation edges: %s, %s", e1, e2)

            # if e1 < allowable_peak_gap or e2 > len(self.original) - allowable_peak_gap:
            #     if logger.isEnabledFor(logging.DEBUG):
            #         logger.debug("Interpolation edges are too close")
            #     continue
            # # Need a check for e2 being too close to the next e1.

            range_1 = np.arange(e1-allowable_peak_gap, e1)
            range_2 = np.arange(e2, e2+allowable_peak_gap)

            if np.any(range_1 <= 0):
                interp_range = [0] + range_2.tolist()
                values = [0] + self.filtered[range_2].tolist()
                poly_fit = np.poly1d(np.polyfit(interp_range, values, poly_order))
            elif np.any(range_2 >= len(mask)-1):
                interp_range = range_1.tolist() + [len(mask)-1]
                values = self.filtered[range_1].tolist() + [0]
                poly_fit = np.poly1d(np.polyfit(interp_range, values, poly_order))
            else:
                interp_range = range_1.tolist() + range_2.tolist()
                poly_fit = np.poly1d(np.polyfit(interp_range,
                                                self.filtered[interp_range],
                                                poly_order))
            self.filtered[e1:e2] = poly_fit(np.arange(e1, e2))

        self.bandpass = signal.savgol_filter(self.filtered,
                                             window_length=window_length,
                                             polyorder=poly_order)
        self.modified = self.original - self.bandpass

==> sslf-0.1.8-py2.py3-none-any.whl: sslf/sslf.py <==
"""
The main module of sslf (simple spectral line finder).
The provided Spectrum class is intended to facilitate all functionality.
"""

# Python 2 and 3 compatibility
from future.builtins import range
from future.utils import raise_with_traceback

import copy
import logging

import numpy as np
import numpy.ma as ma
from scipy import signal


logger = logging.getLogger(__name__)


def find_background_rms(array, num_chunks, use_chunks):
    """
    In order to obtain a better estimate of a spectrum's RMS, break the input
    array into evenly sized chunks (or as even as possible), sort the chunks
    by RMS, then return the average of some of these chunks as the true RMS.

    Parameters
    ----------
    array : array_like
        The 1-dimensional spectrum we wish to determine the RMS for.

    num_chunks : int, optional
        Split `array` into this many chunks for RMS calculation.

    use_chunks : int, optional
        Use this many chunks from `array` for RMS calculation.
    """

    chunks = np.array_split(array, num_chunks)
    if isinstance(array, ma.core.MaskedArray):
        # The following requires np.errstate, because numpy.ma apparently throws
        # floating point errors erroneously.
        # https://github.com/numpy/numpy/issues/4895
        with np.errstate(under="ignore"):
            sorted_by_rms = sorted([np.std(x.data) for x in chunks])
    else:
        sorted_by_rms = sorted([np.std(x) for x in chunks])
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("rms chunks = %s", sorted_by_rms)
    mean = np.mean(sorted_by_rms[:use_chunks])
    if mean == 0:
        raise_with_traceback(ValueError("RMS is 0, which cannot be used."))
    return mean


def _blank_spectrum_part(spectrum, point, radius, value=0):
    lower = max([0, point + int(-radius)])
    upper = min([len(spectrum), point + int(radius)])
    spectrum[lower:upper] = value


class _Peak(object):
    def __init__(self, channel, snr, width):
        self.channel = channel
        self.snr = snr
        self.width = width


class Spectrum(object):
    def __init__(self, spectrum, vel=None, num_chunks=5, use_chunks=3):
        """
        Provide a spectrum to find lines on, and/or remove the bandpass from.
        The optional vel parameter essentially provides an "x-axis" to the
        data, such that peak positions can be determined in terms of this
        axis rather than merely the channel position in the spectrum.
        Note that any NaN values in the spectrum are filtered.

        Parameters
        ----------
        spectrum : array_like
            The 1-dimensional spectrum we will performing line-finding on.

        vel : array_like, optional
            Supply an equally-sized array to `spectrum` to have `sslf` report
            the velocities (or some other arbitrary interpretation of the x-axis
            of `spectrum`) of found lines.

        num_chunks : int, optional
            When calculating the RMS of `spectrum`, break `spectrum` into
            equally-sized chunks according to `num_chunks`.

        use_chunks : int, optional
            When calculating the RMS of `spectrum`, use this many equally-sized
            chunks from when `spectrum` was split according to `num_chunks`.
        """

        # Make sure the data are in numpy arrays.
        if isinstance(spectrum, list):
            spectrum = np.array(spectrum)
        if isinstance(vel, list):
            vel = np.array(vel)

        # Filter any NaNs, scipy doesn't like them.
        if np.any(np.isnan(spectrum)):
            nan_indices = np.isnan(spectrum)
            self.original = spectrum[~nan_indices]
            if vel is not None:
                self.vel = vel[~nan_indices]
            else:
                self.vel = None
        else:
            self.original = spectrum
            self.vel = vel

        self.rms = find_background_rms(spectrum, num_chunks=num_chunks, use_chunks=use_chunks)

    def find_cwt_peaks(self, scales, snr=6.5, wavelet=signal.ricker,
                       num_chunks=5, use_chunks=3):
        """
        From the input spectrum (and a range of scales to search):
        - perform a continuous wavelet transformation (CWT)
        - find a significant peak in the CWT matrix
        - mask this peak in wavelet space, for all scales
        - loop from step 2, until no significant peaks remain
        - return the list of peaks

        In my experience, an SNR of 6.5 is a good compromise for reducing the number
        of false positives found while reliably finding real, significant peaks.

        It may be worthwhile to smooth the spectrum before performing the CWT.

        Parameters
        ----------
        scales : array_like
            The channel-widths to use when performing the CWT, and thus the
            channel-widths we use when searching for spectral lines.

        snr : float, optional
            The significance threshold of any found spectral line.

        wavelet : function
            The wavelet to use when performing the CWT. The wavelet used should
            have a similar profile to any spectral lines that are being
            searched.

        num_chunks : int, optional
            When calculating the RMS of `spectrum`, break `spectrum` into
            equally-sized chunks according to `num_chunks`.

        use_chunks : int, optional
            When calculating the RMS of `spectrum`, use this many equally-sized
            chunks from when `spectrum` was split according to `num_chunks`.
        """

        peaks = []
        cwt_mat = signal.cwt(self.original, wavelet, scales)
        cwt_mat = ma.array(cwt_mat)
        spectrum_length = cwt_mat.shape[1]

        while True:
            peak_pixel = cwt_mat.argmax()
            i, peak_channel = np.unravel_index(peak_pixel, cwt_mat.shape)
            peak = cwt_mat[i, peak_channel]
            rms = find_background_rms(cwt_mat[i], num_chunks=num_chunks, use_chunks=use_chunks)
            sig = peak / rms
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Biggest peak at channel %s, scale %s, rms = %s",
                             peak_channel, scales[i], rms)

            # If this maximum is not significant, we're done.
            if sig < snr:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Peak is not significant (%s < %s); finishing", sig, snr)
                break
            # Otherwise, blank this line across all scales.
            else:
                for j, s in enumerate(scales):
                    # If the line is too close to the edge,
                    # cap the mask at the edge.
                    lower = max([0, peak_channel - 2 * s])
                    upper = min([spectrum_length, peak_channel + 2 * s])
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Blanked channels %s to %s (scale = %s)",
                                     lower, upper, s)
                    cwt_mat[j, lower:upper] = ma.masked
                peaks.append(_Peak(peak_channel, sig, scales[i]))

        self.channel_peaks = [p.channel for p in peaks]
        self.peak_snrs = [p.snr for p in peaks]
        self.peak_widths = [p.width for p in peaks]
        if self.vel is not None:
            self.vel_peaks = [self.vel[p.channel] for p in peaks]
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Channel peaks: %s", self.channel_peaks)
            logger.debug("Peak SNRs: %s", self.peak_snrs)
            logger.debug("Peak widths: %s", self.peak_widths)
            if self.vel is not None:
                logger.debug("Velocity peaks: %s", self.vel_peaks)

    def vel_peaks2chan_peaks(self):
        """
        This function is useful for when you know the velocities of the spectral lines,
        and need to determine the relevant channels before subtracting the bandpass.
        """
        self.channel_peaks = []
        for vp in self.vel_peaks:
            self.channel_peaks.append(np.abs(self.vel - vp).argmin())
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Channel peaks: %s", self.channel_peaks)

    def subtract_bandpass(self,
                          window_length=151,
                          poly_order=1,
                          blank_spectrum_width=1.4,
                          allowable_peak_gap=10,
                          bandpass_func=None):
        """
        Flag the locations of any lines, and subtract the non-zero bandpass
        everywhere else. Provide the flattened spectrum in self.modified.

        Parameters
        ----------
        window_length : int, optional
            The window_length to be used in scipy's savgol_filter. This should
            be bigger than the expected width of any spectral lines, but not
            so big that it is comparable to the length of the spectrum. In a
            sense, it specifies how far to look ahead and behind every channel
            when considering bandpass shape. If this is too small, it will
            behave more like a low-pass filter than a high-pass filter, when we
            probably want it to be more on the high-pass side.

        poly_order : int, optional
            The order of the polynomial fitting to be done over the spectral
            lines and with the savgol_filter. 1 (linear) is generally sensible,
            especially as higher orders are susceptible to Runge's phenomenon.

        blank_spectrum_width : float or callable, optional
            The amount of spectrum to filter on either side of a detected
            spectral line. This is in an effort to preserve real signal
            associated with the spectral line that is buried in the noise.
            If this is a lambda, then a spectral line's SNR is used to calculate
            the number of channels to blank around it. `brewmaster` uses e.g.
            lambda x: 10 * np.sqrt(x) + 3

        allowable_peak_gap : int, optional
            The minimum number of channels between any two spectral lines before
            sslf considers them to be the same for the purposes of filtering.
            This avoids using noisy channels between lines for interpolation,
            and getting a poor subtraction. Currently disabled.

        bandpass_func : callable, optional
            If present, this specifies the function to use when determining a
            bandpass shape for subtraction. The default is
            lambda x: scipy.signal.savgol_filter(x,
                                                 window_length=window_length,
                                                 polyorder=poly_order)
        """

        mask = np.zeros(len(self.original))

        if bandpass_func is None:
            bandpass_func = lambda x: signal.savgol_filter(x,
                                                           window_length=window_length,
                                                           polyorder=poly_order)

        for i, p in enumerate(self.channel_peaks):
            if isinstance(blank_spectrum_width, float):
                width = self.peak_widths[i] * blank_spectrum_width
            elif callable(blank_spectrum_width):
                width = blank_spectrum_width(self.peak_snrs[i])
            else:
                raise_with_traceback(ValueError("blank_spectrum_width is neither a number nor a function."))

            # Blank the lines, fitting the bandpass around them.
            _blank_spectrum_part(mask, p, radius=width, value=1)

        self.filtered = copy.copy(self.original)

        # Find where lines start and end.
        edges = []
        if mask[0] == 1:
            edges.append(0)
        for e in np.where(np.diff(mask))[0]:
            edges.append(e)
        if mask[-1] == 1:
            edges.append(len(mask) - 1)

        # Interpolate between gaps in the spectrum.
        for i in range(len(edges) // 2):
            e1, e2 = edges[2 * i], edges[2 * i + 1]
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Interpolation edges: %s, %s", e1, e2)

            # if e1 < allowable_peak_gap or e2 > len(self.original) - allowable_peak_gap:
            #     if logger.isEnabledFor(logging.DEBUG):
            #         logger.debug("Interpolation edges are too close")
            #     continue
            # # Need a check for e2 being too close to the next e1.

            range_1 = np.arange(e1 - allowable_peak_gap, e1)
            range_2 = np.arange(e2, e2 + allowable_peak_gap)

            if np.any(range_1 <= 0):
                interp_range = [0] + range_2.tolist()
                values = [0] + self.filtered[range_2].tolist()
                poly_fit = np.poly1d(np.polyfit(interp_range, values, poly_order))
            elif np.any(range_2 >= len(mask) - 1):
                interp_range = range_1.tolist() + [len(mask) - 1]
                values = self.filtered[range_1].tolist() + [0]
                poly_fit = np.poly1d(np.polyfit(interp_range, values, poly_order))
            else:
                interp_range = range_1.tolist() + range_2.tolist()
                poly_fit = np.poly1d(np.polyfit(interp_range,
                                                self.filtered[interp_range],
                                                poly_order))
            self.filtered[e1:e2] = poly_fit(np.arange(e1, e2))

        self.bandpass = bandpass_func(self.filtered)
        self.modified = self.original - self.bandpass
